//
// LCD controller for Epson P300013100 display
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Epson P300013100 hardware interface code based on work from Kedric C. Bartsch.
//
// This is a 640 x 200 monochrome display. 4 data bits per column.
// A single Atmega1284P does all the work.
//
// The LCD is internally organized as 100 lines of 1280 bits (320 columns),
// where columns 1..159 occupy the top half of the screen and columns
// 160-319 occupy the bottom half.
// This means 320 nibbles per horizontal line, 100 horizontal lines,
// totalling 16000 bytes, which are all stored in the internal AVR SRAM.
// The screen refresh is done in the Timer0 ISR using a software DotClk (PD7).
// HSYNC (PD5) and VSYNC (PD6) are also generated here, as is the ENABLE (PD4) line.
// There is also a special ECL (PC1) line and an YSCL (PD2) line
//
// Graphics processing is done in the forever loop.
//
// Epson P300013100 LCD specs:
//
// 640 x 1 x 200
// 4-bit parallel data interface.
// No overscan is needed. The LCD actually receives data only about 50% of time (at 60 Hz)
//
// Current Pixel/nibble data are stored in Vpix, Vx and Vy. These are only used for the communication interface.
//
// Internally, these coordinates are converted to physical coordinates inside the memory access functions
//
// We'll use the AVCX Serial Interface Protocol for now, extended and changed for B/W bitmaps.
//
// Bitmaps used are 320 x 240 pixels, where the first and last 20 lines are ignored to fit 320 x 200. Two
// bitmaps will fit on one screen (640 x 200)
//
// Invalid data may cause all kinds of problems, specially out of range errors. The display driver uses Software Flow Control
// on the Serial Port to prevent overflow of the receive buffer.
//
// Once the parser gets out of sync, it may gobble a lot of input bytes if it expects bitmap data.
// Therefore an input timeout is active. The timer countdown value will be reset everytime data is received.
// Once the timer triggers, we should check that the SerialIn buffer is empty (how could it not be) and
// then reset the parser state to idle. Check if there is a little hole when the timer and SerialIn interrupt happen
// almost simultaneously.
//
// Timer1 is running at 100 mSecs. to do LED blinking, inByte time-out, clock update general wait.
//
// Many commands expect the relevant pixel address to be nibble aligned. To enforce this, the nibble address
// is automatically aligned to the closest nibble address before the current pixel address (no change
// if already aligned). The pixel/nibble issue only applies to the X direction.
//
// This display support Dark pixels only (black on "white", actually greenish)
//
// The P300013100 LCD has an oddbal connector which can easily be desoldered, leaving holes for a 20 pin header.
// The definitions of the pins (as published by Kedric C. Bartsch) is a little odd.
//
//		|______________________LCD________________________|
//
//
//				11 12 13 14 15 16 17 18 19 20
//				1  2  3  4  5  6  7  8  9  10
//
//	1:  DB0	2: DB1	3: DB2	4: DB3		(PB0..PB3)
//	5:  Enable							(PD4)
//	6:  AC (or M)						(PD3)
//	7:  YSCL							(PD2)
//	8:  unused
//	9:  ECL								(PC1)
//	10: DotClk							(PD7)
//	11: VCC (+5V)
//	12: GND	13: GND
//	14: VSYNC							(PD6)
//	15: GND
//	16: VEE (-19V app.)
//	17: GND
//  18: HSYNC							(PD5)
//	19: nc.  20: nc
//
// Total: 11 signal lines, 3 power lines
//
//	SED1335 compatibility is only an option if external memory is added.
//
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "./config.h"
#include "./clkisr.h"
#include "./util.h"

#define NIBBLEALIGN(px) if (1) {px = (px + 3) & 0xfffc;}	//align on next 4 bit boundary

// Clock Image Data
#include "./EpsonClk.inc"

extern prog_uint8_t Font8x8[][8];
// SplashScreen 320 x 240 
extern prog_uint8_t SplashScreen[];
//
#define max(a,b) ((a > b) ? a : b)
#define min(a,b) ((a < b) ? a : b)
//
#define COLORON		0b00000001
#define COLOROFF	0b00000000
//
// a cluster is the lowest common denominator of
// pixels and memory units (bytes/nibbles)
// 
#if BITSPERPIXEL == 1
#define nClustersPerLine (nHorPixels / DATABITS)
#elif BITSPERPIXEL == 3
#define nClustersPerLine (nHorPixels / 8)
#endif

#define nLinesPerBar 20
#define CLOCKOFFSET 20			// such that the 200 x 200 Clock and the 80 x 60 picture are centered in the right side window

uint8 FrameBuffer[PHYS_LINES*BYTES_PER_LINE];
uint8 pwrup_state;
uint8 RefreshLine;
uint8 PDATA;
uint8 BDATA;
uint8 Vx, Vy, Vb;			//nibble address, row address, byte address
uint  Vpix;					//pixel address
uint8 ParseState, PrevParseCommand;
uint8 GlobalFault;
uint8 CommandTimeout;
uint  NoActivityTimeout;
uint8 SecsCountdown;
volatile uint8 SecsToAdd;
uint8 TimeSecs, TimeMins, TimeHrs;
uint8 fShowClock, fStartClock, fAutoClock;
uint  XStart, XEnd, XSize;			//current window limits
uint8 YStart, YEnd, YSize;
uint MaxPictureBytes;
// for Line Command
uint8 SavedArgs[9];
uint8 SavedIdx;
unsigned int BMByteCount;
#if EXTDEBUG
uint8 *pSTData;
unsigned long InCnt;
#endif

void LCDDemo(uint8 val);
void Serial_In (uint8 data);
void DrawTable(prog_uint8_t *Table, uint8 erase);

extern void InitMem(void);
#if DEBUG
uint8 ReadNibble(uint8 x, uint8 y)   __attribute__ ((noinline));
void WriteNibble(uint8 val, uint8 x, uint8 y)   __attribute__ ((noinline));
#else
uint8 ReadNibble(uint8 x, uint8 y);
void WriteNibble(uint8 val, uint8 x, uint8 y);
#endif
void SetPixel(uint8 color, uint x, uint8 y);
void PutChar(uint8 ch);

#if DEBUG
void MemoryScanNonZero(uint8 y);
#endif

void Parse(uint8 inByte);

void __attribute__ ((naked, section (".init1")))	// called before .do_copy_data_start
SetMem_init (void) 
{
	InitMem();
}

// __heap_start is defined in the linker script
extern uint8_t __heap_start;

#if DEBUG
//		 Check for 0xcc (in DEBUG mode) as set by SetMem();
uint16_t
GetMemUnused (void)
{
	uint16_t unused = 0;
    
    // Get end of static allocated RAM space (.data, .bss, .noinit, ...)
	uint8_t *p = &__heap_start;
	
	do
	{
        // Value written in InitMem() still intact?
		if (*p++ != 0xcc)
			break;
			
		unused++;
	} while (p <= (uint8_t *) RAMEND);

	return unused;
}
#endif

void foo(void)
{
	do {
		asm volatile (
			"nop	\n\t"
		);
	} while (1);
}
//
// Enable debug led
//
void InitDebug(void)
{
#if USE_LED
	LED_OFF;
	sbi(LED_DDR,LED_BIT);			//output
#endif
	LED2_OFF;
	sbi(LED2_DDR,LED2_BIT);			//output
}

const char DiagStr0[] PROGMEM = "DebugL0 ";
const char DiagStr1[] PROGMEM = "\n\rDebugL1: 0x";
const char DiagStr2[] PROGMEM = "DebugL2\n\r";
const char DiagStr3[] PROGMEM = "DebugL3\n\r";
const char DiagStr4[] PROGMEM = "DebugL4\n\r";
const char DiagStr5[] PROGMEM = "DebugL5\n\r";

void DebugL0(uint8 err)  __attribute__ ((noinline));
void DebugL0(uint8 err)
{
	UARTPrintfProgStr_I((prog_char*)DiagStr0);
	Printfuint8(err);
	PrintfEndOfLine();
	FlashLed(5, 1);
}

void DebugL1(void)
{
	UARTPrintfProgStr_I((prog_char*)DiagStr1);
#if 0
	// Move this to ClkIsr.c
	uint8 *ptr;
	Printfu16((int)TailPtr);
	UARTSendByte(' ', FALSE);
	FlashLed(10, 1);
	//
	// now print last 32 bytes from input queue
	//
	for (int i = 32; i >= 1; --i) {
		ptr = (uint8 *)(TailPtr - i);
		if (ptr < &RxBuffer[0]) ptr += RXSIZE;
		Printfuint8(*ptr);
		UARTSendByte(' ', FALSE);
	}
#endif
	PrintfEndOfLine();
}

void DebugL2(void)
{
	UARTPrintfProgStr_I((prog_char*)DiagStr2);
	FlashLed(10, 2);
}

void DebugL3(void)
{
	UARTPrintfProgStr_I((prog_char*)DiagStr3);
	FlashLed(10, 3);
}

void DebugL4(void)
{
	UARTPrintfProgStr_I((prog_char*)DiagStr4);
	FlashLed(10, 4);
}

void DebugL5(void)
{
	UARTPrintfProgStr_I((prog_char*)DiagStr5);
	FlashLed(10, 5);
}
//
// TC0 is used to do the LCD Screen Refresh
//
// Initialize TC0
// Set prescaler to clkio/64. At 20 Mhz, this is 3.2 uSec
// Using a Timer0CountUp value of 52 yields an match every 166.4 uSec.
// resulting in a refresh rate of 60 (60 x 100 = 6,000 Hz -> 166.66 uSec)
// 60Hz refresh rate shows no visible flicker, a little flicker on video.
//
// Processing time per line: 89 uSec (1/3/2011),
// meaning the LCD refresh load is 53%. At this load, the AVR has no
// trouble processing a 115,200 bps serial stream (plus there is flow control)
//
// The refresh rate can be reduced to 50 Hz and there still is no visible flicker.
//
#if F_CPU==20000000UL
#define Timer0CountUp 52		//62 for 50 Hz.
#else
error: F_CPU undefined
#endif

void InitTC0(void)
{
	TCCR0A = _BV(WGM01);			// run timer in CTC mode
	TCCR0B = (0<<CS02)|_BV(CS01)|_BV(CS00);
	OCR0A = Timer0CountUp;
}

void EnableTC0(void)
{
	TIFR0 = _BV(OCF0A);				//clear pending match
	TIMSK0 = _BV(OCIE0A);			//Enable TC0.match
}

//
// Timer1 is setup in util.c but calls the application
// specific function "AppTimer1Stuff"
//
void AppTimer1Stuff(void)
{
	if (--SecsCountdown == 0) {
		SecsCountdown = 10;
		++SecsToAdd;
	}
	if (CommandTimeout != 0) {
		if (--CommandTimeout == 0) {
			// Command Time out. Reset State Machine
			ParseState = STANDBY;
		}
	}
	if (NoActivityTimeout != 0) {
		if (--NoActivityTimeout == 0) {
			// No Activity Timeout. Start Clock
			// Ought to save current image, then clrscreen first
			fStartClock = TRUE;
			fAutoClock = TRUE;
		}
	}
}

void InitDebug(void);

void InitCtrl(void)
{
//
// I/O PORTS INITIALIZATION
//
 	SRAMDATA_DDR = SRAMDATA_MASK;	//data port defaults to output


	cbi(LCD_YSCL, YSCL_BIT);		//active High
	cbi(LCD_ENABLE, ENABLE_BIT);	//disable LCD
	cbi(LCD_HSYNC, HSYNC_BIT);		//active High
	cbi(LCD_VSYNC, VSYNC_BIT);		//active High
	cbi(DOTCLK, DOTCLK_BIT);		//active on falling edge
	cbi(LCD_ECL, ECL_BIT);			//active High

	sbi(YSCL_DDR, YSCL_BIT);
	sbi(LCDM_DDR, M_BIT);
	sbi(ENABLE_DDR, ENABLE_BIT);
	sbi(HSYNC_DDR, HSYNC_BIT);
	sbi(VSYNC_DDR, VSYNC_BIT);
	sbi(DOTCLK_DDR, DOTCLK_BIT);
	sbi(ECL_DDR, ECL_BIT);

#if DEBUG
	cbi(MCU_TEST, MCU_TEST_BIT);
	sbi(MCU_TEST_DDR, MCU_TEST_BIT);
#endif
//
// Disable unused parts of the AVR to reduce power consumption
//
// unused TWI, ADC, TIM2, SPI
//
// Reserve TIM1 for inByte timeout
//
	PRR0 = _BV(PRTWI) | _BV(PRADC) | _BV(PRTIM2) | _BV(PRSPI);

	InitTC0();					//LCD refresh timer
	InitTC1();					//Timeout/Led
	InitDebug();
//
// VARIABLES INITIALIZATION
//
	XStart = 0;
	XEnd = nHorPixels - 1;
	XSize = XEnd - XStart + 1;
	YStart = 0;
	YEnd = nLines - 1;
	YSize = YEnd - YStart + 1;
	MaxPictureBytes  = (YSize * (XSize / 8));

	RefreshLine = 0;			//LCD Line being displayed
	pwrup_state = 0;
	ParseState = 0;				//GrState = STANDBY
	PrevParseCommand = 0;			//nothing
	PDATA = COLORON;
	BDATA = COLOROFF;
	Vpix = Vx = Vy = 0;
	GlobalFault = 0;
	SecsCountdown = 10;
	SecsToAdd = 0;
	CommandTimeout = 0;
	NoActivityTimeout = NOACTIVITY;
#if EXTDEBUG
	pSTData = FrameBuffer;
	InCnt = 0;
#endif
}
//
// WriteNibble:	write a nibble to SRam.
//
// IN:	r24:nibble	r22 x-address	r20 y-address
//
void WriteNibble(uint8 nibble, uint8 x, uint8 y)
{
	uint8 b, *p;
	uint x0 = x;
	ASSERT(x < LOG_NIBBLES_PER_LINE);
	if (x >= LOG_NIBBLES_PER_LINE) {
		x = LOG_NIBBLES_PER_LINE - 1;
	}
	ASSERT(y < LOG_MAX_LINES);
	if (y >= LOG_MAX_LINES) {
		y = LOG_MAX_LINES - 1;
	}
	if (y >= PHYS_LINES) {
		y -= PHYS_LINES;
		x0 += LOG_NIBBLES_PER_LINE;
	}
	p = &FrameBuffer[(y * BYTES_PER_LINE) + (x0 >> 1)];
	b = *p;
	if (x & 1) {
		b = (nibble & 0x0f) | (b & 0xf0);
	} else {
		b = (nibble << 4) | (b & 0x0f);
	}
	*p = b;
}

//
// WriteByte:	write a byte to SRam.
//
// IN:	r24:byte	r22 x-address (in bytes)	r20 y-address
//
void WriteByte(uint8 b, uint8 x, uint8 y)
{
	ASSERT(x < LOG_NIBBLES_PER_LINE / 2);
	if (x >= (LOG_NIBBLES_PER_LINE /2)) {
		x = (LOG_NIBBLES_PER_LINE / 2) - 1;
	}
	ASSERT(y < LOG_MAX_LINES);
	if (y >= LOG_MAX_LINES) {
		y = LOG_MAX_LINES - 1;
	}
	if (y >= PHYS_LINES) {
		y -= PHYS_LINES;
		x += (LOG_NIBBLES_PER_LINE / 2);
	}
	FrameBuffer[(y * BYTES_PER_LINE) + x] = b;
}

uint8 ReadNibble(uint8 x, uint8 y)
{
	uint8 b;
	uint x0 = x;
	ASSERT(x < LOG_NIBBLES_PER_LINE);
	if (x >= LOG_NIBBLES_PER_LINE) {
		x = LOG_NIBBLES_PER_LINE - 1;
	}
	ASSERT(y < LOG_MAX_LINES);
	if (y >= LOG_MAX_LINES) {
		y = LOG_MAX_LINES - 1;
	}
	if (y >= PHYS_LINES) {
		y -= PHYS_LINES;
		x0 += LOG_NIBBLES_PER_LINE;
	}
	b = FrameBuffer[(y * BYTES_PER_LINE) + (x0 >> 1)];
	if (x & 1) {
		return (b & 0x0f);
	} else {
		return (b >> 4);
	}
}
//
// Write a character at the current pixel location.
// May not wrap Vpix and Vy properly.
// SED
//
void PutChar(uint8 ch)
{
	uint8	i,FontLine,px;
	prog_uint8_t *pCode;
	NIBBLEALIGN(Vpix);
	if ((Vpix + 8) > nHorPixels) {			// char won't fit on current line
		Vpix = 0;
		Vy += 8;							// Line overflow check next
	}
	if ((Vy + 8) > nLines) {				// char won't fit on screen
		Vy = 0;
	}
	if (((uint8)Vpix & 0x07) != 0) {		// not byte aligned
		px = (Vpix >> 2);					// Convert to nibble
		//
		// write One Line at a time
		for (i = 0; i < 8; i++) {
			pCode = &(Font8x8[ch][i]);
			FontLine = __LPM(pCode);
			WriteNibble(FontLine >> 4, px, Vy+i);
			WriteNibble(FontLine & 0x0f, px+1, Vy+i);
		}
	} else {								// Vpix is byte aligned
		px = (Vpix >> 3);					// Convert to byte
		//
		// write One Line at a time
		for (i = 0; i < 8; i++) {
			pCode = &(Font8x8[ch][i]);
			FontLine = __LPM(pCode);
			WriteByte(FontLine, px, Vy+i);
		}
	}
	Vpix += 8;
}

byte MaskTbl[4] PROGMEM = {
	8, 4, 2, 1
};

//
// Set Pixel at location x, y to color
// TODO use 4 byte table to compute bitmask
//
void SetPixel(uint8 color, uint x, uint8 y)
{
	uint8 v, offset, bitmask, px;
	x = min(x, nHorPixels - 1);
	y = min(y, nLines - 1);
	px = (x >> 2);				//enclosing nibble
	v = ReadNibble(px, y);
	offset = x & 0x03;
	bitmask = __LPM(&MaskTbl[offset]);
	v &= ~bitmask;		//clear pixel
	if (color) {
		v |= bitmask;
	}
	WriteNibble(v, px, y);
}
//
// Bresenham line drawing
//
void DrawLine(uint x0, uint y0, uint x1, uint y1)
{
	uint steep, tmp;
	u32 x, y;
	s16 ystep;
	int deltax, deltay, error;

#if DEBUG
	if ((x0 >= nHorPixels) || (x1 >= nHorPixels) || (y0 >= nLines) || (y1 >= nLines)) return;
#endif

	steep = (abs(y1 - y0) > abs (x1 - x0));
	if (steep) {
		tmp = x0; x0 = y0; y0 = tmp;		//  swap(x0, y0)
		tmp = x1; x1 = y1; y1 = tmp;        //  swap(x1, y1)
	}
    if (x0 > x1) {
		tmp = x0; x0 = x1; x1 = tmp;		// swap(x0, x1)
		tmp = y0; y0 = y1; y1 = tmp;		// swap(y0, y1)
	}
	deltax = x1 - x0;
	deltay = abs(y1 - y0);
	error = 0;
	y = y0;
	if (y0 < y1) ystep = 1;
	else ystep = -1;
	for (x = x0; x <= x1; x++) {
		if (steep) SetPixel(PDATA, y, x);
		else SetPixel(PDATA, x, y);
		error += deltay;
		if ((error * 2) >= deltax) {
			y += ystep;
			error -= deltax;
		}
	}
}

//
// Bresenham circle drawing
//
void DrawCircle(unsigned int cx, unsigned int cy ,unsigned int radius)
{
	int x, y, xchange, ychange, radiusError;
	x = radius;
	y = 0;
	xchange = 1 - 2 * radius;
	ychange = 1;
	radiusError = 0;
	while(x >= y) {
	  SetPixel(PDATA, cx+x, cy+y ); 
	  SetPixel(PDATA, cx-x, cy+y ); 
	  SetPixel(PDATA, cx-x, cy-y );
	  SetPixel(PDATA, cx+x, cy-y ); 
	  SetPixel(PDATA, cx+y, cy+x ); 
	  SetPixel(PDATA, cx-y, cy+x ); 
	  SetPixel(PDATA, cx-y, cy-x ); 
	  SetPixel(PDATA, cx+y, cy-x ); 
	  y++;
	  radiusError += ychange;
	  ychange += 2;
	  if ( 2*radiusError + xchange > 0 ) {
	    x--;
		radiusError += xchange;
		xchange += 2;
		}
	}
}

void DrawTable(prog_uint8_t *Table, uint8 erase)
{
	prog_uint8_t *pCode = Table;
	uint8 x0,y0,x1,y1;
	uint8 SavedColor = PDATA;

	if (erase) {
		PDATA = BDATA;
	}

	do {
		while ((x0 = __LPM(pCode)) == 0) {		//skip leading 0 bytes
			++pCode;
		}
		++pCode;
		ASSERT (x0 != 0);
		y0 = min(lastLine, __LPM(pCode));
		y0 = lastLine-y0;						//invert Y
		++pCode;
		do {
			x1 = __LPM(pCode);
			++pCode;
			if (0 == x1) break;				//move to next coordinate pair
			if (0xff == x1) {				//end of coordinate pairs
				if (erase) {
					BDATA = PDATA;
					PDATA = SavedColor;
				}
				return;
			}
			y1 = min(lastLine, __LPM(pCode));
			y1 = lastLine-y1;						//invert Y
			++pCode;
			DrawLine(x0 + CLOCKOFFSET, y0, x1 + CLOCKOFFSET, y1);
			x0 = x1;
			y0 = y1;
		} while (1);
	} while (1);

	if (erase) {
		BDATA = PDATA;
		PDATA = SavedColor;
	}
}

void DrawSecsHand(uint8 sec, uint8 erase)
{
	if (sec >= 60)
		return;

	DrawTable(&SecPtrData[sec * 5], erase);
}

void DrawMinsHand(uint8 min, uint8 erase)
{
	if (min >= 60)
		return;

	DrawTable(&MinPtrData[min * 9], erase);
}

void DrawHrsHand(uint8 hr, uint8 erase)
{
	if (hr >= 12)
		return;

	DrawTable(&HrPtrData[hr * 45], erase);
}

void UpdateTime(void)
{
	while (SecsToAdd > 0) {
		--SecsToAdd;
		if (fShowClock) {
			DrawSecsHand(TimeSecs, TRUE);			//erase current seconds hand
		}
		++TimeSecs;
		if (TimeSecs >= 60) {
			if (fShowClock) {
				DrawMinsHand(TimeMins, TRUE);		//erase current minutes hand
			}
			TimeSecs = 0;
			++TimeMins;
			if (TimeMins >= 60) {
				if (fShowClock) {
					DrawHrsHand(TimeHrs, TRUE);		//erase current hours hand
				}
				TimeMins = 0;
				++TimeHrs;
				if (TimeHrs >= 12) {
					TimeHrs = 0;
				}
			}
		}
		if (fShowClock) {
			DrawTable(DialDigits12, FALSE);
			DrawSecsHand(TimeSecs, FALSE);
			DrawMinsHand(TimeMins, FALSE);
			DrawHrsHand(TimeHrs, FALSE);
		}
	}
}

void StartClock(void)
{
	fStartClock = FALSE;
	fShowClock = TRUE;
	DrawTable(DialData, FALSE);
	DrawTable(DialDigits12, FALSE);
	DrawSecsHand(TimeSecs, FALSE);
	DrawMinsHand(TimeMins, FALSE);
	DrawHrsHand(TimeHrs, FALSE);
}

void StopClock(void)
{
	DrawTable(DialData, TRUE);
	DrawTable(DialDigits12, TRUE);
	DrawSecsHand(TimeSecs, TRUE);
	DrawMinsHand(TimeMins, TRUE);
	DrawHrsHand(TimeHrs, TRUE);
	fShowClock = FALSE;
	fAutoClock = FALSE;
}

uint8 PrepareNibble(uint8 color)
{
#if DEBUG
	color &= 0x01;
#endif
	return (color) ? 0b00001111 : 0;
}
//
// set screen area to color. Set BDATA, Vx, Vy, Vpix
//
void SetScreenX(uint8 color, uint x0, uint8 y0, uint x1, uint8 y1)
{
	uint8 nibble, x, y;
//
// convert pixel address X0 to nibble boundary
// align starting pixel to next nibble boundary
// align ending pixel to previous nibble boundary
//
	NIBBLEALIGN(x0);
	x1 = (x1 + 0) & 0xfffc;

	nibble = PrepareNibble(color);
	for (y = y0; y <= y1; ++y) {		// Top to Bottom
		for (x = (x0 >> 2) ; x <= (x1 >> 2); ++x) {
			WriteNibble(nibble, x, y);
		}
	}
	BDATA = color;						//set color as background color
	Vy = y0;
	Vx = (x0 >> 2);
	Vpix = (Vx << 2);
}

//
// set screen to color. Set BDATA, Vx, Vy, Vpix
//
// Optimization: could to a memset on the FrameBuffer!
//
void SetScreen(uint8 color)
{
	uint8 v, x, y;
#if DEBUG
	color &= 0x01;
#endif
	v = (color) ? 0b11111111 : 0;
	for (y = 0; y < nLines; ++y) {		// Top to Bottom
		for (x = 0 ; x < (LOG_NIBBLES_PER_LINE / 2); ++x) {
			WriteByte(v, x, y);
		}
	}
	BDATA = color;						//set color as background color
	Vy = Vx = Vpix = 0;
}

#if DEBUG
//
// Draw a line using argument color. Starting at current Vy
// Use WriteNibble()
//
void HorizontalLine(uint8 color)
{
	uint8 x, nibble;
	nibble = PrepareNibble(color);
	for (x = 0; x < nColumnsPerLine; ++x) {
		WriteNibble(nibble, x, Vy);
	}
}

//
// Draw a line using argument color. Starting at current Vy.
// Use SetPixel
//
void HorizontalPixelLine(uint8 color)
{
	PDATA = color;
	uint px;
	for (px = 0; px < (nHorPixels/16); ++px) {
		SetPixel(color, px, Vy);
	}
}
#endif

//
// Draw a bar using argument color. Starting at current Vy
//
void HorizontalBar(uint8 color)
{
	uint8 i, x, nibble;
	i = nLinesPerBar;			//number of lines
	nibble = PrepareNibble(color);
	do {
		for (x = 0; x < nColumnsPerLine; ++x) {
			WriteNibble(nibble, x, Vy);
		}
		++Vy;
	} while (--i);
	Vx = 0;
}
//	
// horizontal test bars.
//
void HorizontalTest(void)
{
	uint8 i;
	Vy = 0;
	for (i = 0; i < (nLines/(2 * nLinesPerBar)); ++i) {		//as many as will fit
		HorizontalBar(COLORON);
		HorizontalBar(COLOROFF);
	}
#if DEBUG
	if (Vy != nLines) {
		DebugL0(LCDERROR0);
	}
#endif
	Vx = Vy = 0;
}
//
//  compare data with PDATA
//
void VerifyNibble(uint8 data)
{
	// bit 0: D0 .. bit 3: D0
	if ((data & 0x01) != PDATA) DebugL0(LCDERROR0);
	data >>= 1;
	if ((data & 0x01) != PDATA) DebugL0(LCDERROR0);
	data >>= 1;
	if ((data & 0x01) != PDATA) DebugL0(LCDERROR0);
	data >>= 1;
	if ((data & 0x01) != PDATA) DebugL0(LCDERROR0);
}

void VerifyHorizontalLine(void)
{
	uint8 nibble, i = nClustersPerLine;
	Vx = 0;
	do {
		nibble = ReadNibble(Vx, Vy);
		++Vx;
		VerifyNibble(nibble);		//compare with PDATA
	} while (--i);
}
//
// argument: expected color 
//
void VerifyHorizontalBar(uint8 color)
{
	uint8 i = nLinesPerBar;
	PDATA = color;			//save
	do {
		VerifyHorizontalLine();
		++Vy;
	} while (--i);
}

void VerifyHorizontalTest()
{
	Vy = 0;
	for (uint8 i = 0; i < (nLines/(2 * nLinesPerBar)); ++i) {		//as many as will fit
		VerifyHorizontalBar(COLORON);
		VerifyHorizontalBar(COLOROFF);
	}
	Vx = Vy = 0;				//reset
	Vpix = 0;
}
//
// VerticalTest: draw vertical bars
//
void VerticalTest()
{
	uint8 color, i, nibble, x, y;
	for (y = 0; y < nLines; ++y) {
		x = 0;
		color = COLOROFF;		//color
		for (i = 0; i < (nColumnsPerLine >> 1); ++i) {
			if (++color > COLORON) {	//next color
				color = COLOROFF;
			}
			nibble = PrepareNibble(color);	//set from PDATA
			WriteNibble(nibble, x, y);
			WriteNibble(nibble, x+1, y);
			x += 2;
		}
	}
	Vx = Vy = 0;				//reset
	Vpix = 0;
}

//
// HorVertTest: draw clusters
//
void HorVertTest()
{
	uint8 i, j, color0, color, nibble, x, y;
	color0 = COLOROFF;				//start color
	for (y = 0; y < nLines;) {
		if (++color0 > COLORON) {	//next color
			color0 = COLOROFF;
		}
		color = color0;
		for (j = 0; j < 8; ++j) {		//nr of lines per block
			x = 0;
			for (i = 0; i < (nColumnsPerLine >> 1); ++i) {
				nibble = PrepareNibble(color);	//set  from PDATA
				WriteNibble(nibble, x, y);
				WriteNibble(nibble, x+1, y);
				x += 2;
				if (++color > COLORON) {	//next color
					color = COLOROFF;
				}
			}
			++y;
		}
	}
	Vx = Vy = 0;			//reset
	Vpix = 0;
}
//
// Draw horizontal bars slowly.
//
void SlowHorizontalBars(void)
{
	uint8 i;
	Vy = 0;
	for (i = 0; i < (nLines/(2 * nLinesPerBar)); ++i) {		//as many as will fit
		HorizontalBar(COLORON);
		HorizontalBar(COLOROFF);
		delayOneSecond();
	}
	Vx = 0;
	Vy = 0;
}
//
// Select a demo. All delays done by sender
//
void LCDDemo(uint8 val)
{
	if (val == 1) {
		HorizontalTest();
		VerifyHorizontalTest();
	} else if (val == 2) {
		VerticalTest();
	} else if (val == 3) {
		HorVertTest();
	} else if (val == 4) {
		SlowHorizontalBars();		//takes a while
	}
}

void Parse(uint8 inByte)
{
	uint8 v;
	//
	// Setting CommandTimeout to 50 gives us 5 second
	// before the ParseState is set to STANDBY
	// Timeout will never occur while inside Parse() - that's the intent
	//
	CommandTimeout = 50;
	if (ParseState == STANDBY) {
		//
		//Decode inByte
		//
  		if ((inByte & SINGLEBYTEMASK) != SINGLEBYTECMD) {
			//
			//MultiByteCommand
			//
			PrevParseCommand = inByte;
			ParseState = inByte & MULTIBYTEMASK;
		}
		else {
			//
			// SingleByteCommand
			//
			if (LCD_DUMMY == inByte) {
			} else if (CLEAR_SCR == inByte) {
				SetScreen(COLOROFF);		// sets BDATA
				PDATA = COLORON;
			} else if (CLOCK_ON == inByte) {
				StartClock();
			} else if (CLOCK_OFF == inByte) {
				StopClock();
			}
			//
			// done with unparameterized single byte commands.
			// Must be <PSET>, <X++>, <Y++>, <X-->, <Y-->, <X++,PSET>, <Y++,PSET>, <X--,PSET> or, <Y--,PSET>
			// Also UNSET (Clr pixel)
			else if ((inByte & PSETINCDECMASK) == PSETINCDECCMD) {	//Increment, Decrement, PSET
				if ((inByte & PSETPIXMASK) == PSETPIXMASK) {		// unused
				} else if (inByte & PSETPIXON) {
					SetPixel(PDATA, Vpix, Vy);			// Set Pixel
				} else if (inByte & PSETPIXOFF) {
					SetPixel(BDATA, Vpix, Vy);			//Clear Pixel
				}
				//
				// Now check for X/Y increment or decrement
				//
				v = (inByte & PSETXYMASK);				//X, Y or nothing
				if (v == PSETXYMASK) {					//unused
				} else if (v == PSETX) {
					if (inByte & PSETDEC) {				//Decrement
						if (--Vpix == 0xffff) Vpix = lastHorPixel;
					} else {							//Increment
						if (++Vpix == nHorPixels) Vpix = 0;
					}
				} else if (v == PSETY) {
					if (inByte & PSETDEC) {				//Decrement
						if (--Vy == 0xff) Vy = lastLine;
					} else {							//Increment
						if (++Vy == nLines) Vy = 0;
					}
				} else {								//No X or Y action.
				}
			} //PSETINCDECCMD
			else { // UNKNOWN or UNIMPLEMENTED
			//ignore since it's single byte
			}
		} // SingleByteCommand
	} //ParseState == STANDBY
#if EXTDEBUG
	else if (ParseState == SERIALTEST) {
  		if ((PrevParseCommand & MULTIBYTEMASK) == SERIALTEST) {
			//First time through.
			SavedIdx = 0;
			PrevParseCommand &= ~MULTIBYTEMASK;
		}
		// 1 byte argument: #of test bytes
		SavedArgs[SavedIdx++] = inByte;
		if (SavedIdx >= 1) {
			ParseState = SERIALTEST_DATA;
			PrevParseCommand = SERIALTEST;
		}
	}
	else if (ParseState == SERIALTEST_DATA) {
		if (PrevParseCommand != SERIALTEST) {
			// inconsistency
				DebugL1();
		}
		*pSTData++ = inByte;		//store test bytes
		if (pSTData >= &FrameBuffer[PHYS_LINES*BYTES_PER_LINE]) {
			pSTData = FrameBuffer;
		}
		if (--SavedArgs[0] == 0) {
			ParseState = STANDBY;
		}
	}
#endif
	else if (ParseState == PICTURE) {
		//
		// TODO. sb pixel argument for x. If x0 or x1 not aligned
		// on nibble boundary, discard the data.
		// first 4 byte arguments (x0, y0, x1, y1. Nibble addresses
		// then data: (((x1 - x0 + 1) * (y1 - y0 + 1)) / 2) bytes
  		if ((PrevParseCommand & MULTIBYTEMASK) == PICTURE) {
			//First time through.
			SavedIdx = 0;
			PrevParseCommand &= ~MULTIBYTEMASK;
		}
		// 4 bytes argument: X0,Y0,X1,Y1 (nibble addresses)
		SavedArgs[SavedIdx++] = inByte;
		if (SavedIdx >= 4) {	// collected all args
			BMByteCount = ((SavedArgs[2] - SavedArgs[0] + 1) * (SavedArgs[3] - SavedArgs[1] + 1) / 2);
			if (BMByteCount > MaxPictureBytes) {
				// TODO. Compute excess and ignore those bytes
				DebugL1();
			}
			// TODO: expect Pixel Argument? Cluster Argument. Use X window limits
			if (SavedArgs[0] >= nClustersPerLine) SavedArgs[0] = 0;		//TODO
			Vx = SavedArgs[0];
//			if (SavedArgs[1] < YStart) SavedArgs[1] = YStart;
			if (SavedArgs[3] > YEnd) SavedArgs[3] = YEnd;
			Vy = SavedArgs[3];
			ParseState = PICTURE_DATA;
#if DEBUG
			PrevParseCommand = PICTURE;
#endif
		}
	}	// PICTURE
	else if (ParseState == PICTURE_DATA) {
		ASSERT(PrevParseCommand == PICTURE);
		// 2 nibbles packed per byte, packed Upper nibble, then Lower nibble
		// works as long as "x1 - x0 + 1" is even
		// Data come in bottom of area lines first
		if (((uint8)Vx & 0x01) != 0) {		// not byte aligned
			WriteNibble((inByte >> 4), Vx, Vy);
			WriteNibble((inByte & 0x0f), Vx+1, Vy);
		} else {
			WriteByte(inByte, Vx >> 1, Vy);
		}
		Vx += 2;
		if ((Vx - 1) == SavedArgs[2]) {
			Vx = SavedArgs[0];
			Vpix = (Vx << 2);
			if (--Vy == SavedArgs[1] - 1) {
				Vy = SavedArgs[1];
			}
		}
		if (--BMByteCount == 0) {
			ParseState = STANDBY;
		}

	} //PICTURE_DATA
	else if ((ParseState & MULTIBYTEMASK) == COLOR) {
		PDATA = (PrevParseCommand & 0x07);
		BDATA = PDATA ? 0 : 1;
		// ignore current byte in inByte
		ParseState = STANDBY;
	}
	else if (ParseState == SET) {
  		if ((PrevParseCommand & MULTIBYTEMASK) == SET) {
			//First time through.
			SavedIdx = 0;
			PrevParseCommand &= ~MULTIBYTEMASK;
		}
		// 2 bytes argument: X or Y
		SavedArgs[SavedIdx++] = inByte;
		if (SavedIdx >= 2) {	// collected all args
			if (PrevParseCommand & SETX) Vpix = SavedArgs[0] | (SavedArgs[1] << 8);
			else Vy = (uint8)(SavedArgs[0] | (SavedArgs[1] << 8));
			if (PrevParseCommand & SETPIXON) {
				SetPixel(PDATA, Vpix, Vy);
			} else if (PrevParseCommand & SETPIXOFF) {
				SetPixel(BDATA, Vpix, Vy);
			}
			ParseState = STANDBY;
		}
	}
	else if (ParseState == PRINT_STRING) {
		//
		// Print String State. Check if NULL
		//
		if (inByte != 0) {
			PutChar(inByte);
		}
		else {
			ParseState = STANDBY;	// End of String
		}
	}
	else if (ParseState == PUT_CHAR) {
		PutChar(inByte);
		ParseState = STANDBY;
	}
	else if (ParseState == DRAW_LINE) {
  		if ((PrevParseCommand & MULTIBYTEMASK) == DRAW_LINE) {
			//First time through.
			SavedIdx = 0;
			PrevParseCommand &= ~MULTIBYTEMASK;
		}
		// 8 bytes argument: X0,Y0,X1,Y1
		// X0, X1, Y0 and Y1 are 2 bytes each (low, high)
		SavedArgs[SavedIdx++] = inByte;
		if (SavedIdx >= 8) {	// collected all args
			DrawLine(SavedArgs[0] | (SavedArgs[1] << 8),
					 SavedArgs[2] | (SavedArgs[3] << 8),
					 SavedArgs[4] | (SavedArgs[5] << 8),
					 SavedArgs[6] | (SavedArgs[7] << 8));
			//
			// Set current coordinate to last pixel drawn
			//
			Vpix = SavedArgs[4] | (SavedArgs[5] << 8);
			Vy = SavedArgs[6] | (SavedArgs[7] << 8);
			ParseState = STANDBY;
		}
	}
	else if (ParseState == DRAW_CIRCLE) {
  		if ((PrevParseCommand & MULTIBYTEMASK) == DRAW_CIRCLE) {
			//First time through.
			SavedIdx = 0;
			PrevParseCommand &= ~MULTIBYTEMASK;
		}
		// 6 bytes argument: X,Y,Radius
		// X, Y and Radius are 2 bytes each (low, high)
		SavedArgs[SavedIdx++] = inByte;
		if (SavedIdx >= 6) {	// collected all args
			DrawCircle(SavedArgs[0] | (SavedArgs[1] << 8),
					   SavedArgs[2] | (SavedArgs[3] << 8),
					   SavedArgs[4] | (SavedArgs[5] << 8));
			//
			// Set current coordinate to center of circle
			//
			Vpix = SavedArgs[0] | (SavedArgs[1] << 8);
			Vy = SavedArgs[2] | (SavedArgs[3] << 8);
			ParseState = STANDBY;
		}
	}
	else if (ParseState == SET_SCREEN) {
  		if ((PrevParseCommand & MULTIBYTEMASK) == SET_SCREEN) {
			//First time through.
			SavedIdx = 0;
			PrevParseCommand &= ~MULTIBYTEMASK;
		}
		// 9 bytes argument: color, X0,Y0,X1,Y1
		// color is 1 byte. X0, X1, Y0 and Y1 are 2 bytes each (low, high)
		SavedArgs[SavedIdx++] = inByte;
		if (SavedIdx >= 9) {	// collected all args
			SetScreenX(	SavedArgs[0],
						SavedArgs[1] | (SavedArgs[2] << 8),				//x0
						(uint8)(SavedArgs[3] | (SavedArgs[4] << 8)),	//y0
						SavedArgs[5] | (SavedArgs[6] << 8),				//x1
						(uint8)(SavedArgs[7] | (SavedArgs[8] << 8)));	//y1
			ParseState = STANDBY;
		}
	}
	else if (ParseState == LCDDEMO) {
		LCDDemo(inByte);		//select Demo
		ParseState = STANDBY;
	}
	else {
		DebugL1();
		ParseState = STANDBY;
	}
#if 0&DEBUG
	if (STANDBY == ParseState) {
		UARTSendByte('.', FALSE);
	}
#endif
}

const char StartStr1[] PROGMEM = "\r\nEpson P3000 LCD Controller V1.2\r\n";
//
// TODO: replace columnOffset with SetWindow()
void SetSplashScreen(uint8 columnOffset)
{
//
// should compress image before enabling this feature
//
// Bitmap is for QVGA (320 x 240) screen.
// Skip first and last 20 lines (6400 pixels, 800 bytes)
//
	prog_uint8_t *pCode = SplashScreen;
	uint8 x, v;
	int y;			//needed for for loop
	pCode += 800;
	for (y = YEnd; y >= YStart; --y) {
		//OK to use WriteByte since horizontal offset X is always even.
		for (x = 0; x < (nColumnsPerLine / 2); x += 2) {
				v = __LPM(pCode++);
				WriteByte(v, (x / 2) + (columnOffset / 2), y);
		}
	}
	fStartClock = TRUE;	//TEST TODO
}

void Preset(void)
{
	Serial_In(CLEAR_SCR);		//clear Screen
	Serial_In(PRINT_STRING);	//String
  	Serial_In('R');
	Serial_In('e');
	Serial_In('a');
	Serial_In('d');
	Serial_In('y');
	Serial_In(0);
}

int main(void)
{
	unsigned long i = 0;
	uint8 inByte;
	MCUSR = 0;
	GPIOR0 = 0;
	InitCtrl();
	if (0 == UARTInit(BAUD_115200)) {
		FlashLed(10, 100);
	}

	fShowClock = FALSE;
	fStartClock = FALSE;
	fAutoClock = FALSE;
	TimeSecs = 0;
	TimeMins = 0;
	TimeHrs = 1;
	SetScreen(COLOROFF);		// sets BDATA
	PDATA = COLORON;
#if SPLASHSCREEN
	SetSplashScreen(0);
	SetSplashScreen(80);
#else
	Preset();		// text to lines 0..7 through SerialIn()
#endif
//
//	Enable Timer0 & Timer1 in Interrupt Mask
//
	EnableTC0();
	EnableTC1();
	sei();							//start interrupts -> screen refresh, serial in
	BlinkFast();
	UARTPrintfProgStr((prog_char*)StartStr1);

#if 0
	SetScreen(COLOROFF);		// sets BDATA
	PDATA = COLORON;
	DrawTable(DialData, FALSE);
	DrawTable(DialDigits12, FALSE);
	DrawSecsHand(TimeMins, FALSE);
	DrawMinsHand(TimeSecs, FALSE);
	DrawHrsHand(TimeHrs, FALSE);
#endif

	do {
		//
		// process queued graphic commands from SerialIn
		//
		while (UARTRXAvailable()) {
			inByte = UARTReceiveByte();
			if (fAutoClock) {
				StopClock();
			}
			Parse(inByte);
			NoActivityTimeout = NOACTIVITY;		// 1 minute
		}

		if (GlobalFault) {
			DebugL0(GlobalFault);
		}
		//
		// Signal when buffer is empty.
		if (++i == 0x4ffff) {
			FlashLed(1, 5);
			i = 0;
			ASSERT(GetMemUnused() > 25);
		}

		if (fStartClock &!fShowClock) {
			SetScreen(COLOROFF);		// sets BDATA
			PDATA = COLORON;
			SetSplashScreen(80);
			StartClock();
		}
		fStartClock = FALSE;

		if (SecsToAdd > 0) {
			UpdateTime();
		}
	} while (TRUE);
}


